#include <iostream>
#include <algorithm>
#include <vector>
#include <string>
#include <cstdio>
#include <climits>
#include <cstdlib>
#include <ctime>
#include <utility>
#include <sstream>
#include <ctime>
#include <cmath>
#ifdef IMG
#include "EasyBMP/EasyBMP.h"
#endif	// IMG
using namespace std;

typedef unsigned int ui;
typedef unsigned long long int ulli;
typedef unsigned long int uli;
typedef long double ld;
typedef unsigned char uc;

#define DIM 200

struct Point
{
	uc X;
	uc Y;

	Point()
	{
		X = Y = 0;
	}

	Point(uc x, uc y)
	{
		X = x;
		Y = y;
	}

	void print()
	{
		cout << "("
			 << (ui) X
			 << ", "
			 << (ui) Y
			 << ")";
	}

	static Point* getRandom(ui delta = 1)
	{
		uc x, y;
		x = (uc) (rand() % (DIM - delta));	// [0, 199 - delta]
		y = (uc) (rand() % (DIM - delta));	// [0, 199 - delta]
		return new Point(x, y);
	}
};

struct Rectangle
{
	Point* TL;
	Point* BR;
	uc H;
	uc W;

	Rectangle(Point* tl, Point* br)
	{
		TL = tl;
		BR = br;
		H = BR->Y - TL->Y;
		W = BR->X - TL->X;
	}

	~Rectangle()
	{
		delete TL;
		delete BR;
	}

	/*
	static uc randCoordinate(int t = -1)
	{
		uc ret;
		do
		{
			ret = (uc) (rand() % DIM);
		}
		while(ret <= t);
		return ret;
	}
	*/

	void print()
	{
		cout << "Rectangle: ";
		TL->print();
		cout << " --> ";
		BR->print();
		cout << "\t"
			 << (ui) W
			 << " x "
			 << (ui) H
			 << " = "
			 << ((ui) W) * H
			 << endl;
	}

	static Rectangle* getRandom()	// TODO: implement corresponding function\
											 to return a random rectangle with\
											 equal H and W.
	{
		uc x1, x2, y1, y2;
		x1 = (uc) (rand() % (DIM - 1));		// [0, 198]
		y1 = (uc) (rand() % (DIM - 1));		// [0, 198]
		x2 = (uc) (rand() % (DIM - x1));	// [0, 200 - x1 - 1]
		x2 = x1 + (x2 ? x2 : 1);			// 200 > x2 > x1
		y2 = (uc) (rand() % (DIM - y1));	// [0, 200 - y1 - 1]
		y2 = y1 + (y2 ? y2 : 1);			// 200 > y2 > y1

		return new Rectangle(new Point(x1, y1), new Point(x2, y2));
	}

	#define POINTGAPDIFFRECTANGLES			15
	#define POINTGAPSAMERECTANGLES			1
	inline static uc MIN(uc a, uc b)
	{
		return (a > b ? b : a);
	}
	inline static uc ABSDIFF(uc a, uc b)
	{
		return (a > b ? (a - b) : (b - a));
	}

	static pair<Rectangle*, Rectangle*> getRandom2()
	{
		Point *p1, *p2, *p1_, *p2_;
		uc temp1, temp2;

		p1_ = new Point();
		p2_ = new Point();

		getrandompoints:
		p1 = Point::getRandom();
		p2 = Point::getRandom();
		
		if(ABSDIFF(p1->X, p2->X) < POINTGAPDIFFRECTANGLES || ABSDIFF(p1->Y, p2->Y) < POINTGAPDIFFRECTANGLES)
		{
			delete p1;
			delete p2;
			goto getrandompoints;
		}

		temp1 = (uc) (rand() % ABSDIFF(p1->X, p2->X));
		temp1 = temp1 ? temp1 : 1;
		temp2 = MIN(rand() % (DIM - p1->X), rand() % (DIM - p2->X));
		temp2 = temp2 ? temp2 : 1;
		temp1 = temp2 = MIN(temp1, temp2);
		if(temp1 < POINTGAPSAMERECTANGLES)
		{
			delete p1;
			delete p2;
			goto getrandompoints;
		}

		p1_->X = p1->X + temp1;
		p2_->X = p2->X + temp2;
		
		temp1 = (uc) (rand() % ABSDIFF(p1->Y, p2->Y));
		temp1 = temp1 ? temp1 : 1;
		temp2 = MIN(rand() % (DIM - p1->Y), rand() % (DIM - p2->Y));
		temp2 = temp2 ? temp2 : 1;
		temp1 = temp2 = MIN(temp1, temp2);
		if(temp1 < POINTGAPSAMERECTANGLES)
		{
			delete p1;
			delete p2;
			goto getrandompoints;
		}

		p1_->Y = p1->Y + temp1;
		p2_->Y = p2->Y + temp2;
		
		return make_pair(new Rectangle(p1, p1_), new Rectangle(p2, p2_));
	}

	bool areOverlapping(Rectangle*& that) const
	{
		if(		this->TL->X < that->BR->X
			&&	this->BR->X > that->TL->X
			&&	this->TL->Y < that->BR->Y
			&&	this->BR->Y > that->TL->Y)
		{
			return true;
		}
		return false;
	}
	
};

struct Pixel
{
	uc R;
	uc G;
	uc B;

	Pixel()
	{
		clear();
	}

	Pixel(uc r, uc g, uc b)
	{
		R = r;
		G = g;
		B = b;
	}

	void clear()
	{
		R = G = B = 0;
	}

	void update(Pixel*& that)
	{
		this->R = that->R;
		this->G = that->G;
		this->B = that->B;
	}

	inline static ulli SQDI(ulli a, ulli b)
	{
		ulli sqr = a - b;
		return sqr * sqr;
	}

	ulli distance(Pixel& that)
	{
		return (ulli) sqrt(	SQDI(this->R, that.R)
						+	SQDI(this->G, that.G)
						+	SQDI(this->B, that.B) );
	}

	inline static uc AVG(ui a, ui b)
	{
		ui avg = a + b;
		return (uc) (avg / 2);
	}
	
	Pixel* average(Pixel& that)
	{
		return new Pixel(	AVG(this->R, that.R),
							AVG(this->G, that.G),
							AVG(this->B, that.B)	);
	}
};

struct Image
{
	Pixel Matrix[DIM][DIM];

	Image()
	{
		/*
		Matrix = (Pixel**) malloc(DIM * sizeof(Pixel*));
		for(int i = 0; i < DIM; i++)
		{
			Matrix[i] = (Pixel*) malloc(DIM * sizeof(Pixel));
		}
		*/
	}

	void read()
	{
		for(int i = 0; i < DIM; i++)
			for(int j = 0; j < DIM; j++)
				scanf("%hhu", &(Matrix[i][j].R));
		for(int i = 0; i < DIM; i++)
			for(int j = 0; j < DIM; j++)
				scanf("%hhu", &(Matrix[i][j].G));
		for(int i = 0; i < DIM; i++)
			for(int j = 0; j < DIM; j++)
				scanf("%hhu", &(Matrix[i][j].B));
	}

	#ifdef IMG
	void copyToBMPImage(BMP& bmp)
	{
		for(int i = 0; i < DIM; i++)
		{
			for(int j = 0; j < DIM; j++)
			{
				bmp(i, j)->Red = Matrix[i][j].R;
				bmp(i, j)->Green = Matrix[i][j].G;
				bmp(i, j)->Blue = Matrix[i][j].B;
			}
		}
	}
	#endif	// IMG

	ulli distance(Image*& that)
	{
		ulli d = 0;
		for(ui i = 0; i < DIM; i++)
			for(ui j = 0; j < DIM; j++)
				d += (this->Matrix[i][j]).distance(that->Matrix[i][j]);
		return d;
	}

	pair<ulli, ulli> distancesInRectangle(Image*& that, Rectangle*& r1, Rectangle*& r2)
	{
		ulli d1, d2;
		d1 = d2 = 0;
		
		Pixel *pAveragePixel;
		ui x1, y1, x2, y2;
		for(ui i = 0; i < r1->W; i++)
		{
			x1 = r1->TL->X + i;
			x2 = r2->TL->X + i;
			
			for(ui j = 0; j < r1->H; j++)
			{
				y1 = r1->TL->Y + j;
				y2 = r2->TL->Y + j;

				// TODO: optimize; compute and store values; modify distance as well.
				d1 += (this->Matrix[x1][y1]).distance(that->Matrix[x1][y1]);
				pAveragePixel = (this->Matrix[x1][y1]).average(this->Matrix[x2][y2]);
				d2 += pAveragePixel->distance(that->Matrix[x1][y1]);
				delete pAveragePixel;
			}
		}

		return make_pair(d1, d2);
	}

	void averageInRectangle(Rectangle*& r1, Rectangle*& r2)
	{
		// TODO: optimize using cached values from the above function
		Pixel *pAveragePixel;
		ui x1, y1, x2, y2;
		for(ui i = 0; i < r1->W; i++)
		{
			x1 = r1->TL->X + i;
			x2 = r2->TL->X + i;
			
			for(ui j = 0; j < r1->H; j++)
			{
				y1 = r1->TL->Y + j;
				y2 = r2->TL->Y + j;

				pAveragePixel = (this->Matrix[x1][y1]).average(this->Matrix[x2][y2]);
				(this->Matrix[x1][y1]).update(pAveragePixel);
				delete pAveragePixel;
			}
		}
	}
};

ui K;
Image* Source;
Image* Target;
#ifdef IMG
BMP SourceBMP;
BMP TargetBMP;
#endif	// IMG

ui totalmorpharea;
ui successfulmorphs;
ui totalmorphs;

struct OutputElement
{
  uc x1, x2, y1, y2, h, w;

  OutputElement()
  {
    x1 = x2 = y1 = y2 = h = w = UCHAR_MAX;
  }

  OutputElement(uc xx1, uc yy1, uc xx2, uc yy2, uc hh, uc ww)
  {
    x1 = xx1;
    x2 = xx2;
    y1 = yy1;
    y2 = yy2;
    h = hh;
    w = ww;
  }
};
vector<OutputElement*> output;

bool morph();

#ifdef IMG
void writeMorphBMP(ui i)
{
	stringstream filename;
	filename << "images/morph_" 
			 << i
			 << ".bmp";

	Source->copyToBMPImage(SourceBMP);
	SourceBMP.WriteToFile(filename.str().c_str());
	
}
#endif	// IMG

#ifdef DBG
#define ENDLOOPTIME	58
#endif	// DBG
#ifndef DBG
#define ENDLOOPTIME	14
#endif	// !DBG
#define MORPHAREAUB	4000000

int main(void)
{
	clock_t timekeeper = clock();
	srand(8484);

	freopen("transformation.in", "r", stdin);
	freopen("transformation.out", "w", stdout);
	
	scanf("%u", &K);	// is actually H
	scanf("%u", &K);	// is actually W
	scanf("%u", &K);

	Source = new Image();
	Source->read();

	Target = new Image();
	Target->read();

	#ifdef IMG
	SourceBMP.SetSize(DIM, DIM);
	TargetBMP.SetSize(DIM, DIM);

	#ifdef DBG
	Source->copyToBMPImage(SourceBMP);
	SourceBMP.WriteToFile("images/SourceBMP.bmp");

	Target->copyToBMPImage(TargetBMP);
	TargetBMP.WriteToFile("images/TargetBMP.bmp");
	#endif	// DBG
	#endif	// IMG

	output.clear();
	totalmorpharea = 0;
	totalmorphs = 0;
	successfulmorphs = 0;
	while(	successfulmorphs < K
		&&	totalmorpharea <= MORPHAREAUB
		&&	(((ld) clock() - timekeeper) / CLOCKS_PER_SEC) < ENDLOOPTIME)
	{
		successfulmorphs += morph() ? 1 : 0;
		totalmorphs++;
		#ifdef IMG
		#ifdef DBG
		if(!(successfulmorphs % 100))
		{
			writeMorphBMP(successfulmorphs);
		}
		#endif	// DBG
		#endif	// IMG
	}

	#ifdef IMG
	writeMorphBMP(UINT_MAX);
	#endif	// IMG

	#ifdef DBG
	cout << "Successful morph()s = "
		 << successfulmorphs
		 << "/"
		 << totalmorphs
		 << " = "
		 << ((float) successfulmorphs * 100) / totalmorphs
		 << "%"
		 << endl;

	cout << "Total morph()ed area = "
		 << totalmorpharea
		 << "/"
		 << MORPHAREAUB
		 << " = "
		 << ((float) totalmorpharea * 100) / MORPHAREAUB
		 << "%"
		 << endl;

	cout << "Total distance between Source and Target = "
		 << Source->distance(Target)
		 << endl;
	#endif	// DBG

	#ifndef DBG
	printf("%lu\n", output.size());
	for(ui i = 0; i < output.size(); i++)
		printf(	"%hhu %hhu %hhu %hhu %hhu %hhu\n",
				output[i]->x1,
				output[i]->y1,
				output[i]->x2,
				output[i]->y2,
				output[i]->w,
				output[i]->h );
	#endif	// !DBG

	return 0;
}

#define BASENUFFNESS	3000
inline bool isEnoughDistance(ulli d)
{
	return (d > ((ulli) (((ld) BASENUFFNESS * (1 + successfulmorphs)) / (1 + totalmorphs))));
}

bool morph()
{
	// Find two rectangles with random endpoints and equal areas.
	pair<Rectangle*, Rectangle*> rects = Rectangle::getRandom2();
	#ifdef DBG
	rects.first->print();
	rects.second->print();
	cout << "Are Overlapping? "
		 << rects.first->areOverlapping(rects.second)
		 << endl;
	/*
	Rectangle* rect1 = Rectangle::getRandom();
	Rectangle* rect2 = Rectangle::getRandom();
	rect1->print();
	rect2->print();
	cout << "Are Overlapping? "
		 << rect1->areOverlapping(rect2)
		 << endl
		 << endl;
	*/
	#endif	// DBG

	// Calculate distance deltas.
	pair<ulli, ulli> distances = Source->distancesInRectangle(Target, rects.first, rects.second);
	#ifdef DBG
	cout << "d1 = "
		 << distances.first
		 << ", "
		 << "d2 = "
		 << distances.second
		 << endl;
	#endif	// DBG

	if(distances.first < distances.second || !isEnoughDistance(distances.first - distances.second))
	{
		#ifdef DBG
		cout << "DISCARDING RECTANGLES!!!" << endl << endl;
		#endif	// DBG

		delete rects.first;
		delete rects.second;
		return false;
	}
	else
	{
		#ifdef DBG
		cout << "Applied Morph." << endl << endl;
		#endif	// DBG
	
		totalmorpharea += ((ui) rects.first->H) * rects.first->W;

		#ifndef DBG
		output.push_back(new OutputElement(	rects.first->TL->X,
											rects.first->TL->Y,
											rects.second->TL->X,
											rects.second->TL->Y,
											rects.first->H,
											rects.first->W ));
		#endif	// !DBG
	}
	//cout << endl;

	// Maximize rectangles if distance keeps on decreasing.
	// TODO

	// Apply the operation to source image.
	Source->averageInRectangle(rects.first, rects.second);

	// Store the applied transformation, checking criteria.

	delete rects.first;
	delete rects.second;
	return true;
}
